'' =================================================================================================
''
''   File....... thric_parser.spin 
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2011 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 23 DEC 2011
''
'' =================================================================================================


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000

  LOOP_MS  = 25                                                 ' main loop timing


con

  RX1 = 31
  TX1 = 30
  SDA = 29
  SCL = 28


con

  BUF_SIZE = 80

  CMD_HDR  = ">"                                                ' command string header
  FLD_SEP  = ","                                                ' field separator
  EOM      = 13                                                 ' end of message (carriage return)
  BS       =  8

  P_COUNT  = 8                                                  ' allow eight parameters


con

   #1, HOME, GOTOXY, #8, BKSP, TAB, LF, CLREOL, CLRDN, CR       ' PST formmatting control
  #14, GOTOX, GOTOY, CLS


obj

  term : "fullduplexserial64"
  s    : "jm_strings"
 

var

  long  cmd
  long  nparams
  long  param[P_COUNT]

  byte  bidx
  byte  inbuf[BUF_SIZE]


pub main | t, c

  setup

  term.str(string(CLS, "Parser Test", CR, CR))

  t := cnt
  repeat
  
    c := term.rxcheck                                           ' anything in serial buffer?
    if (c => 0)                                                 ' if yes
      enqueue(c)                                                '  add to input buffer

      ' show input status

      term.tx(HOME)
      term.tx(LF)
      term.tx(LF)      
      term.str(@inbuf)                                           ' show buffer
      term.tx(CLREOL)
      term.tx(CR)
      term.dec(bidx)                                            ' show current buffer index
      term.tx(" ")
      term.tx(CR)
      term.tx(CLRDN)

      ' was last input EOM and new command available?

      if (cmd > 0)
        show_cmd
        reset

  
    waitcnt(t += constant(LOOP_MS * MS_001))                    ' run loop at fixed rate


pub setup

  term.start(RX1, TX1, %0000, 115_200)                          ' start serial io for terminal
  pause(1)

  reset                                                         ' reset command parser  


pub show_cmd | idx

  term.str(string("-- Command...... "))
  term.dec(cmd)
  term.tx(CR)

  term.str(string("-- Parameters... "))
  term.dec(nparams)
  term.tx(CR)

  if (nparams > 0)
    repeat idx from 0 to nparams-1
      term.str(string("    "))
      term.dec(param[idx])
      term.tx(CR)

  term.tx(CR)  


con

pub reset

'' Reset command elements to no command

  cmd := 0                                                      ' invalid cmd #
  nparams := 0                                                  ' no parameters
  longfill(@param, 0, P_COUNT)                                  ' clear parameters
  
  bidx := 0                                                     ' start of buffer                     
  bytefill(@inbuf, 0, BUF_SIZE)                                 ' clear buffer


pub enqueue(c)

'' Add c to input buffer
'' -- if EOM detected, parse command

  c := s.ucase(c)                                               ' convert to uppercase

  if (bidx == 0)                                                ' at beginning?
    if (c == CMD_HDR)                                           '  proper header?
      inbuf[bidx++] := CMD_HDR                                  '   add to buffer
    else
      return                                                    '   ignore input

  else
    case c
      CMD_HDR:                                                  ' (re)start of string?
        reset
        inbuf[bidx++] := CMD_HDR                

      BS:                                                       ' backspace
        inbuf[--bidx] := 0                                      '  clear old, back up

      EOM:                                                      ' end of message?
        inbuf[bidx] := 0                                        '  terminate as z-string                                       
        parse                                                   '  parse command & parameters

      other:
        inbuf[bidx++] := c                                      ' add c to buffer, adance index  

  return cmd
  
  
pub parse | idx, fp

'' Parse command and parameters from input buffer

  cmd := 0
  nparams := 0
  longfill(@param, 0, P_COUNT)

  ' look for known headers
  ' -- string is at position 1 in buffer (just after header character)

  if (s.instr(@inbuf, string("WLO")) == 1)
    cmd := 1

  if (s.instr(@inbuf, string("WLA")) == 1)
    cmd := 2    

  if (s.instr(@inbuf, string("DGS")) == 1)
    cmd := 3

  if (cmd == 0)
    return


  ' parse out parameters

  nparams := count_char(FLD_SEP, @inbuf) <# P_COUNT             ' should be one separator per parameter

  if (nparams > 0)                                              ' parameters in string?
    repeat idx from 1 to nparams                                ' loop through them
      fp := field_pos(@inbuf, idx, FLD_SEP)                     ' find position in buffer
      param[idx-1] := s.asc2val(@inbuf[fp])                     ' extract value


pub count_char(c, spntr) | count, b

'' Counts occurrences of c in string at spntr

  count := 0
  
  repeat
    b := byte[spntr++]
    if (b == 0)
      quit
    elseif (b == c)
      ++count

  return count


pub field_pos(spntr, fn, c) | idx 

'' Returns position index of field number, fn (0+) in string at spntr
'' -- c is the field separator character

  if (fn < 0)                                                   ' must be positive
    return -1

  idx := 0                                                      ' start at beginning

  repeat while (fn > 0)                                         ' loop through field count
    case byte[spntr][idx++]                                     ' get character from string
      c:                                                        ' if separator
        --fn                                                    '  decrement field
                                                                 
      EOM, 0:                                                   ' if end-of-message
        idx := -1                                               '  bad field   
        quit

  return idx


con     

pub pause(ms) | t

'' Delay program ms milliseconds

  t := cnt - 1088                                               ' sync with system counter
  repeat (ms #> 0)                                              ' delay must be > 0
    waitcnt(t += MS_001)
    

pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles pin state

  !outa[pin]
  dira[pin] := 1


pub input(pin)

'' Makes pin input and returns current state

  dira[pin] := 0

  return ina[pin]


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}